import {
  Box,
  styled,
  Table,
  TableBody,
  TableCell,
  TableContainer,
  TableHead,
  TableRow,
  Typography
} from "@mui/material"
import {CSSTransition, TransitionGroup} from 'react-transition-group'
import {useQueries, useQuery} from "@tanstack/react-query";
import {useCallback, useEffect, useMemo, useRef} from "react";

const StyledTableRow = styled(TableRow)(({ theme, tabletype }) => ({
  'tbody &:nth-of-type(odd)': {
    backgroundColor: theme.palette.sncf[tabletype].light,
  },
  'th, &:nth-of-type(even)': {
    backgroundColor: theme.palette.sncf[tabletype].dark,
  },
  // hide last border
  '&:last-child td, &:last-child th': {
    border: 0,
  },
}));

function TrainsTable({stop, date, time, tableType}) {
  return <>
    <TableContainer>
      <Table>
        <TrainsTableHeader tableType={tableType} />
        <TrainsTableBody stop={stop} date={date} time={time} tableType={tableType} />
      </Table>
    </TableContainer>
  </>
}

function TrainsTableHeader({tableType}) {
  return <>
    <TableHead>
      <StyledTableRow tabletype={tableType}>
        <TableCell colSpan="2" fontSize={16} fontWeight="bold">Train</TableCell>
        <TableCell fontSize={16} fontWeight="bold">Heure</TableCell>
        <TableCell fontSize={16} fontWeight="bold">Destination</TableCell>
      </StyledTableRow>
    </TableHead>
  </>
}

function TrainsTableBody({stop, date, time, tableType}) {
  const filterTime = useCallback((train) => {
    if (tableType === "departures")
      return `${train.departure_date}T${train.departure_time_24h}` >= `${date}T${time}`
    else
      return `${train.arrival_date}T${train.arrival_time_24h}` >= `${date}T${time}`
  }, [date, time, tableType])

  const updateTrains = useCallback(() => {
    return fetch(`/api/station/next_${tableType}/?stop_id=${stop.id}&date=${date}&time=${time}&offset=${0}&limit=${20}`)
        .then(response => response.json())
        .then(data => data.results)
        .then(data => [...data])
  }, [stop.id, date, time, tableType])

  const trainsQuery = useQuery({
    queryKey: ['trains', stop.id, tableType],
    queryFn: updateTrains,
    enabled: !!stop.id,
  })
  const trains = useMemo(() => trainsQuery.data ?? [], [trainsQuery.data])

  useEffect(() => {
    let validTrains = trains?.filter(filterTime) ?? []
    if (trains?.length > 0 && validTrains.length < trains?.length)
      trainsQuery.refetch().then()
  }, [trains, filterTime, trainsQuery])

  const nullRef = useRef(null)
  let table_rows = trains.map((train) => <CSSTransition key={train.id} timeout={500} classNames="shrink" nodeRef={nullRef}>
      <TrainRow train={train} tableType={tableType} date={date} time={time} />
    </CSSTransition>)

  return <>
    <TableBody>
      <TransitionGroup component={null}>
        {table_rows}
      </TransitionGroup>
    </TableBody>
  </>
}

function TrainRow({train, tableType, date, time}) {
  const tripQuery = useQuery({
    queryKey: ['trip', train.trip],
    queryFn: () => fetch(`/api/gtfs/trip/${train.trip}/`)
        .then(response => response.json()),
    enabled: !!train.trip,
  })
  const trip = tripQuery.data ?? {}

  const routeQuery = useQuery({
    queryKey: ['route', trip.route],
    queryFn: () => fetch(`/api/gtfs/route/${trip.route}/`)
        .then(response => response.json()),
    enabled: !!trip.route,
  })
  const route = routeQuery.data ?? {}
  const trainType = getTrainType(train, route)

  const stopTimesQuery = useQuery({
    queryKey: ['stop_times', trip.id],
    queryFn: () => fetch(`/api/gtfs/stop_time/?trip=${trip.id}&order=stop_sequence&limit=1000`)
        .then(response => response.json())
        .then(data => data.results),
    enabled: !!trip.id,
  })
  const stopTimes = stopTimesQuery.data ?? []
  const stopIds = stopTimes.map(stop_time => stop_time.stop)

  const stopQueries = useQueries({
    queries: stopIds.map(stopId => ({
      queryKey: ['stop', stopId],
      queryFn: () => fetch(`/api/gtfs/stop/${stopId}/`)
        .then(response => response.json()),
      enabled: !!stopId,
    })),
  })
  const stops = stopTimes.map(((stopTime, i) => ({...stopTime, stop: stopQueries[i]?.data ?? {"name": "…"}}))) ?? []

  let headline = stops[tableType === "departures" ? stops.length - 1 : 0]?.stop ?? {name: "Chargement…"}

  const realtimeTripQuery = useQuery({
    queryKey: ['realtimeTrip', trip.id, date, time],
    queryFn: () => fetch(`/api/gtfs-rt/trip_update/${trip.id}/`)
        .then(response => response.json()),
    enabled: !!trip.id,
  })
  const realtimeTripData = realtimeTripQuery.data ?? {}
  const scheduleRelationship = realtimeTripData.schedule_relationship ?? 0

  const realtimeQuery = useQuery({
    queryKey: ['realtime', train.id, date, time],
    queryFn: () => fetch(`/api/gtfs-rt/stop_time_update/${train.id}/`)
        .then(response => response.json()),
    enabled: !!train.id,
  })
  const realtimeData = realtimeQuery.data ?? {}

  const delay = tableType === "departures" ? realtimeData.departure_delay : realtimeData.arrival_delay
  const prettyDelay = delay && scheduleRelationship !== 3 ? getPrettyDelay(delay) : ""
  const [prettyScheduleRelationship, scheduleRelationshipColor] = getPrettyScheduleRelationship(scheduleRelationship)
  console.log(realtimeTripData)

  let stopsFilter
  if (scheduleRelationship === 3)
    stopsFilter = (stop_time) => true
  else if (tableType === "departures")
    stopsFilter = (stop_time) => stop_time.stop_sequence > train.stop_sequence && stop_time.drop_off_type === 0
  else
    stopsFilter = (stop_time) => stop_time.stop_sequence < train.stop_sequence && stop_time.pickup_type === 0
  let stopsNames = stops.filter(stopsFilter).map(stopTime => stopTime?.stop.name ?? "").join(", ") ?? ""

  return <>
    <StyledTableRow tabletype={tableType}>
      <TableCell>
        <div>
          <Box display="flex"
               justifyContent="center"
               alignItems="center"
               textAlign="center"
               width="4em"
               height="4em"
               borderRadius="15%"
               fontWeight="bold"
               backgroundColor={`#${getBackgroundColor(train, route)}`}
               color={`#${getTextColor(train, route)}`}>
            {trainType}
          </Box>
        </div>
      </TableCell>
      <TableCell>
        <Box display="flex" alignItems="center" justifyContent="center" textAlign="center">
          <div>
            <div>{trip.short_name}</div>
            <div>{trip.headsign}</div>
          </div>
        </Box>
      </TableCell>
      <TableCell>
        <Box display="flex" alignItems="center" justifyContent="center">
          <Box>
            <Box fontWeight="bold" color="#FFED02" fontSize={24}>
              {getDisplayTime(train, tableType)}
            </Box>
            <Box color={delay && delay !== "00:00:00" ? "#e86d2b" : "white"} fontWeight={delay && delay !== "00:00:00" ? "bold" : ""}>
              {prettyDelay}
            </Box>
            <Box color={scheduleRelationshipColor} fontWeight="bold">
              {prettyScheduleRelationship}
            </Box>
          </Box>
        </Box>
      </TableCell>
      <TableCell>
        <Typography fontSize={24} fontWeight="bold" data-stop-id={headline.id}>{headline.name}</Typography>
        <span className="stops">{stopsNames}</span>
      </TableCell>
    </StyledTableRow>
  </>
}

function getTrainType(train, route) {
  if (train.id.startsWith("IDFM"))
    return route.short_name
  else {
    let trainType = train.stop.split("StopPoint:OCE")[1].split("-")[0]
    if (trainType === "Train TER")
      trainType = "TER"
    else if (trainType === "INTERCITES")
      trainType = "INTER-CITÉS"
    else if (trainType === "INTERCITES de nuit")
      trainType = "INTER-CITÉS de nuit"
    return trainType
  }
}

function getBackgroundColor(train, route) {
  if (route.color)
    return route.color
  else if (getTrainType(train, route) === "OUIGO")
    return "E60075"
  return "FFFFFF"
}

function getTextColor(train, route) {
  if (route.text_color)
    return route.text_color
  else {
    let trainType = getTrainType(train, route)
    switch (trainType) {
      case "OUIGO":
        return "FFFFFF"
      case "TGV INOUI":
        return "9B2743"
      case "ICE":
        return "B4B4B4"
      case "INTER-CITÉS":
      case "INTER-CITÉS de nuit":
        return "404042"
      default:
        return "000000"
    }
  }
}

function getDisplayTime(train, tableType) {
  let time = tableType === "departures" ? train.departure_time : train.arrival_time
  let day_split = time.split(' ')
  return day_split[day_split.length - 1].substring(0, 5)
}

function getPrettyDelay(delay) {
  let delay_split = delay.split(':')
  let hours = parseInt(delay_split[0])
  let minutes = parseInt(delay_split[1])
  let full_minutes = hours * 60 + minutes
  return full_minutes ? `+${full_minutes} min` : "À l'heure"
}

function getPrettyScheduleRelationship(scheduledRelationship) {
  switch (scheduledRelationship) {
    case 1:
      return ["Ajouté", "#3ebb18"]
    case 3:
      return ["Supprimé", "#ff8701"]
    default:
      return ["", ""]
  }
}

export default TrainsTable;